@@ -13,6 +13,11 @@ pub use self::error::{Result, RocflError};
 pub use self::repo::OcflRepo;
 pub use self::store::layout::{LayoutExtensionName, StorageLayout};
 pub use self::types::*;
+pub use self::validate::{
+    ErrorCode, IncrementalValidator, IncrementalValidatorImpl, ObjectValidationResult,
+    ProblemLocation, StorageValidationResult, ValidationError, ValidationResult, ValidationWarning,
+    WarnCode,
+};
 
 mod bimap;
 mod consts;
@@ -22,7 +27,9 @@ mod inventory;
 mod lock;
 mod paths;
 mod repo;
+mod serde;
 mod specs;
 mod store;
 mod types;
 mod util;
+mod validate;
