@@ -1,5 +1,6 @@
 //! Local filesystem OCFL storage implementation.
 
+use std::borrow::Cow;
 use std::cell::RefCell;
 use std::collections::HashMap;
 use std::convert::TryFrom;
@@ -24,7 +25,9 @@ use super::{OcflLayout, OcflStore, StagingStore};
 use crate::ocfl::consts::*;
 use crate::ocfl::error::{not_found, Result, RocflError};
 use crate::ocfl::inventory::Inventory;
-use crate::ocfl::{paths, specs, util, InventoryPath, VersionNum};
+use crate::ocfl::store::{Listing, Storage};
+use crate::ocfl::validate::{IncrementalValidator, ObjectValidationResult, Validator};
+use crate::ocfl::{paths, specs, util, ContentPath, InventoryPath, LogicalPath, VersionRef};
 
 static OBJECT_ID_MATCHER: Lazy<RegexMatcher> =
     Lazy::new(|| RegexMatcher::new(r#""id"\s*:\s*"([^"]+)""#).unwrap());
@@ -38,9 +41,14 @@ pub struct FsOcflStore {
     // TODO this never expires entries and is only intended to be useful within the scope of the cli
     /// Caches object ID to path mappings
     id_path_cache: RwLock<HashMap<String, String>>,
+    validator: Validator<FsStorage>,
     closed: Arc<AtomicBool>,
 }
 
+pub struct FsStorage {
+    storage_root: PathBuf,
+}
+
 impl FsOcflStore {
     /// Creates a new FsOcflStore
     pub fn new(storage_root: impl AsRef<Path>) -> Result<Self> {
@@ -63,6 +71,7 @@ impl FsOcflStore {
         let storage_layout = load_storage_layout(&storage_root);
 
         Ok(Self {
+            validator: Validator::new(FsStorage::new(storage_root.clone())),
             storage_root,
             storage_layout,
             id_path_cache: RwLock::new(HashMap::new()),
@@ -77,6 +86,7 @@ impl FsOcflStore {
         init_new_repo(&root, layout.as_ref())?;
 
         Ok(Self {
+            validator: Validator::new(FsStorage::new(root.clone())),
             storage_root: root,
             storage_layout: layout,
             id_path_cache: RwLock::new(HashMap::new()),
@@ -137,16 +147,22 @@ impl FsOcflStore {
         );
 
         let mut iter =
-            InventoryIter::new_id_matching(&self.storage_root, &object_id, self.closed.clone())?;
+            InventoryIter::new_id_matching(&self.storage_root, object_id, self.closed.clone())?;
 
-        match iter.next() {
-            Some(inventory) => {
-                if let Ok(mut cache) = self.id_path_cache.write() {
-                    cache.insert(object_id.to_string(), inventory.object_root.clone());
+        loop {
+            return match iter.next() {
+                Some(Ok(inventory)) => {
+                    if let Ok(mut cache) = self.id_path_cache.write() {
+                        cache.insert(object_id.to_string(), inventory.object_root.clone());
+                    }
+                    Ok(inventory)
                 }
-                Ok(inventory)
-            }
-            None => Err(not_found(&object_id, None)),
+                Some(Err(e)) => {
+                    error!("{:#}", e);
+                    continue;
+                }
+                None => Err(not_found(object_id, None)),
+            };
         }
     }
 
@@ -169,7 +185,7 @@ impl FsOcflStore {
                 Ok(inventory)
             }
         } else {
-            Err(not_found(&object_id, None))
+            Err(not_found(object_id, None))
         }
     }
 
@@ -226,7 +242,7 @@ impl OcflStore for FsOcflStore {
 
         match self.get_object_root_path(object_id) {
             Some(object_root) => self.get_inventory_by_path(object_id, &object_root),
-            None => self.scan_for_inventory(&object_id),
+            None => self.scan_for_inventory(object_id),
         }
     }
 
@@ -236,7 +252,7 @@ impl OcflStore for FsOcflStore {
     fn iter_inventories<'a>(
         &'a self,
         filter_glob: Option<&str>,
-    ) -> Result<Box<dyn Iterator<Item = Inventory> + 'a>> {
+    ) -> Result<Box<dyn Iterator<Item = Result<Inventory>> + 'a>> {
         self.ensure_open()?;
 
         Ok(Box::new(match filter_glob {
@@ -253,8 +269,8 @@ impl OcflStore for FsOcflStore {
     fn get_object_file(
         &self,
         object_id: &str,
-        path: &InventoryPath,
-        version_num: Option<VersionNum>,
+        path: &LogicalPath,
+        version_num: VersionRef,
         sink: &mut dyn Write,
     ) -> Result<()> {
         self.ensure_open()?;
@@ -263,7 +279,7 @@ impl OcflStore for FsOcflStore {
 
         let content_path = inventory.content_path_for_logical_path(path, version_num)?;
         let mut storage_path = PathBuf::from(&inventory.storage_path);
-        storage_path.push(content_path.as_ref().as_ref());
+        storage_path.push(content_path.as_path());
 
         let mut file = File::open(storage_path)?;
         io::copy(&mut file, sink)?;
@@ -360,7 +376,7 @@ impl OcflStore for FsOcflStore {
 
         fs::rename(version_path, &destination)?;
 
-        if let Err(e) = self.copy_inventory_files(&inventory, &destination, &object_root) {
+        if let Err(e) = self.copy_inventory_files(inventory, &destination, &object_root) {
             if let Err(e) = fs::rename(&destination, version_path) {
                 error!("Failed to rollback version {} of object {} at {}: {}. Manual intervention may be required.",
                        version_str, inventory.id, version_path.to_string_lossy(), e);
@@ -439,10 +455,48 @@ impl OcflStore for FsOcflStore {
         Ok(extensions)
     }
 
+    /// Validates the specified object and returns any problems found. Err will only be returned
+    /// if a non-validation problem was encountered.
+    fn validate_object(
+        &self,
+        object_id: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        let object_root = self.lookup_or_find_object_root_path(object_id)?;
+
+        self.validator
+            .validate_object(Some(object_id), &object_root, fixity_check)
+    }
+
+    /// Validates the specified object at the specified path, relative the storage root, and
+    /// returns any problems found. Err will only be returned if a non-validation problem was
+    /// encountered.
+    fn validate_object_at(
+        &self,
+        object_root: &str,
+        fixity_check: bool,
+    ) -> Result<ObjectValidationResult> {
+        self.validator
+            .validate_object(None, object_root, fixity_check)
+    }
+
+    /// Validates the structure of an OCFL repository as well as all of the objects in the repository
+    /// When `fixity_check` is `false`, then the digests of object content files are not validated.
+    ///
+    /// The storage root is validated immediately, and an incremental validator is returned that
+    /// is used to lazily validate the rest of the repository.
+    fn validate_repo<'a>(
+        &'a self,
+        fixity_check: bool,
+    ) -> Result<Box<dyn IncrementalValidator + 'a>> {
+        Ok(Box::new(self.validator.validate_repo(fixity_check)?))
+    }
+
     /// Instructs the store to gracefully stop any in-flight work and not accept any additional
     /// requests.
     fn close(&self) {
         self.closed.store(true, Ordering::Release);
+        self.validator.close();
     }
 }
 
@@ -478,7 +532,7 @@ impl StagingStore for FsOcflStore {
             .open(paths::object_namaste_path(&storage_path))?;
 
         writeln!(file, "{}", OCFL_OBJECT_VERSION)?;
-        self.stage_inventory(&inventory, false, false)?;
+        self.stage_inventory(inventory, false, false)?;
 
         Ok(())
     }
@@ -488,12 +542,12 @@ impl StagingStore for FsOcflStore {
         &self,
         inventory: &Inventory,
         source: &mut impl Read,
-        logical_path: &InventoryPath,
+        logical_path: &LogicalPath,
     ) -> Result<()> {
-        let content_path = inventory.new_content_path_head(&logical_path)?;
+        let content_path = inventory.new_content_path(logical_path);
 
         let mut storage_path = PathBuf::from(&inventory.storage_path);
-        storage_path.push(&content_path.as_ref());
+        storage_path.push(content_path.as_path());
 
         fs::create_dir_all(storage_path.parent().unwrap())?;
         io::copy(source, &mut File::create(&storage_path)?)?;
@@ -505,15 +559,15 @@ impl StagingStore for FsOcflStore {
     fn copy_staged_file(
         &self,
         inventory: &Inventory,
-        src_content: &InventoryPath,
-        dst_logical: &InventoryPath,
+        src_content: &ContentPath,
+        dst_logical: &LogicalPath,
     ) -> Result<()> {
         let object_root = PathBuf::from(&inventory.storage_path);
 
-        let dst_content = inventory.new_content_path_head(&dst_logical)?;
+        let dst_content = inventory.new_content_path(dst_logical);
 
-        let src_storage = object_root.join(src_content.as_ref());
-        let dst_storage = object_root.join(dst_content.as_ref());
+        let src_storage = object_root.join(src_content.as_path());
+        let dst_storage = object_root.join(dst_content.as_path());
 
         fs::create_dir_all(dst_storage.parent().unwrap())?;
         fs::copy(&src_storage, &dst_storage)?;
@@ -526,12 +580,12 @@ impl StagingStore for FsOcflStore {
         &self,
         inventory: &Inventory,
         source: &impl AsRef<Path>,
-        logical_path: &InventoryPath,
+        logical_path: &LogicalPath,
     ) -> Result<()> {
-        let content_path = inventory.new_content_path_head(&logical_path)?;
+        let content_path = inventory.new_content_path(logical_path);
 
         let mut storage_path = PathBuf::from(&inventory.storage_path);
-        storage_path.push(&content_path.as_ref());
+        storage_path.push(content_path.as_path());
 
         fs::create_dir_all(storage_path.parent().unwrap())?;
         fs::rename(source, &storage_path)?;
@@ -543,15 +597,15 @@ impl StagingStore for FsOcflStore {
     fn move_staged_file(
         &self,
         inventory: &Inventory,
-        src_content: &InventoryPath,
-        dst_logical: &InventoryPath,
+        src_content: &ContentPath,
+        dst_logical: &LogicalPath,
     ) -> Result<()> {
         let object_root = PathBuf::from(&inventory.storage_path);
 
-        let dst_content = inventory.new_content_path_head(&dst_logical)?;
+        let dst_content = inventory.new_content_path(dst_logical);
 
-        let src_storage = object_root.join(src_content.as_ref());
-        let dst_storage = object_root.join(dst_content.as_ref());
+        let src_storage = object_root.join(src_content.as_path());
+        let dst_storage = object_root.join(dst_content.as_path());
 
         fs::create_dir_all(dst_storage.parent().unwrap())?;
         fs::rename(&src_storage, &dst_storage)?;
@@ -561,11 +615,11 @@ impl StagingStore for FsOcflStore {
     }
 
     /// Deletes staged content files.
-    fn rm_staged_files(&self, inventory: &Inventory, paths: &[&InventoryPath]) -> Result<()> {
+    fn rm_staged_files(&self, inventory: &Inventory, paths: &[&ContentPath]) -> Result<()> {
         let object_root = PathBuf::from(&inventory.storage_path);
 
         for path in paths.iter() {
-            let full_path = object_root.join(path.as_ref());
+            let full_path = object_root.join(path.as_path());
             info!("Deleting staged file: {}", full_path.to_string_lossy());
             util::remove_file_ignore_not_found(&full_path)?;
             util::clean_dirs_up(full_path.parent().unwrap())?;
@@ -590,7 +644,7 @@ impl StagingStore for FsOcflStore {
                         .to_string();
                     let content_path = util::convert_backslash_to_forward(&relative);
                     if !inventory
-                        .contains_content_path(&InventoryPath::try_from(content_path.as_ref())?)
+                        .contains_content_path(&ContentPath::try_from(content_path.as_ref())?)
                     {
                         info!("Deleting orphaned file: {}", file.path().to_string_lossy());
                         util::remove_file_ignore_not_found(file.path())?;
@@ -633,7 +687,7 @@ impl StagingStore for FsOcflStore {
         if finalize {
             let version_path = paths::version_path(&object_root, inventory.head);
             fs::create_dir_all(&version_path)?;
-            self.copy_inventory_files(&inventory, &object_root, &version_path)?;
+            self.copy_inventory_files(inventory, &object_root, &version_path)?;
         }
 
         Ok(())
@@ -690,23 +744,27 @@ impl InventoryIter {
         })
     }
 
-    fn create_if_matches<P: AsRef<Path>>(&self, object_root: P) -> Option<Inventory> {
+    fn create_if_matches<P: AsRef<Path>>(&self, object_root: P) -> Option<Result<Inventory>> {
         let inventory_path = paths::inventory_path(&object_root);
 
         if self.id_matcher.is_some() {
-            if let Some(object_id) = self.extract_object_id(&inventory_path) {
-                if self.id_matcher.as_ref().unwrap().deref()(&object_id) {
-                    return parse_inventory_optional(object_root, &self.root);
+            match self.extract_object_id(&inventory_path) {
+                Some(Ok(object_id)) => {
+                    if self.id_matcher.as_ref().unwrap().deref()(&object_id) {
+                        Some(parse_inventory(object_root, &self.root))
+                    } else {
+                        None
+                    }
                 }
+                Some(Err(e)) => Some(Err(e)),
+                None => None,
             }
         } else {
-            return parse_inventory_optional(object_root, &self.root);
+            Some(parse_inventory(object_root, &self.root))
         }
-
-        None
     }
 
-    fn extract_object_id<P: AsRef<Path>>(&self, path: P) -> Option<String> {
+    fn extract_object_id<P: AsRef<Path>>(&self, path: P) -> Option<Result<String>> {
         let mut matches: Vec<String> = vec![];
 
         let result = Searcher::new().search_path(
@@ -721,29 +779,25 @@ impl InventoryIter {
         );
 
         if let Err(e) = result {
-            error!(
+            Some(Err(RocflError::General(format!(
                 "Failed to locate object ID in inventory at {}: {:#}",
-                path.as_ref().to_string_lossy(),
+                path.as_ref().display(),
                 e
-            );
-            None
+            ))))
         } else {
             match matches.get(0) {
-                Some(id) => Some(id.to_string()),
-                None => {
-                    error!(
-                        "Failed to locate object ID in inventory at {}",
-                        path.as_ref().to_string_lossy()
-                    );
-                    None
-                }
+                Some(id) => Some(Ok(id.to_string())),
+                None => Some(Err(RocflError::General(format!(
+                    "Failed to locate object ID in inventory at {}",
+                    path.as_ref().display()
+                )))),
             }
         }
     }
 }
 
 impl Iterator for InventoryIter {
-    type Item = Inventory;
+    type Item = Result<Inventory>;
 
     fn next(&mut self) -> Option<Self::Item> {
         loop {
@@ -764,42 +818,106 @@ impl Iterator for InventoryIter {
                 None => {
                     self.current.replace(None);
                 }
-                Some(Err(e)) => error!("{:#}", e),
-                Some(Ok(entry)) => match entry.file_type() {
-                    Err(e) => error!("{:#}", e),
-                    Ok(ftype) if ftype.is_dir() => {
+                Some(Err(e)) => return Some(Err(e.into())),
+                Some(Ok(entry)) => {
+                    let ftype = match entry.file_type() {
+                        Ok(ftype) => ftype,
+                        Err(e) => return Some(Err(e.into())),
+                    };
+
+                    if ftype.is_dir() {
                         let path = entry.path();
 
                         if path.file_name().unwrap_or_default() == EXTENSIONS_DIR {
                             continue;
                         }
 
-                        match is_object_root(&path) {
-                            Ok(is_root) if is_root => {
-                                if let Some(inventory) = self.create_if_matches(&path) {
-                                    return Some(inventory);
-                                }
+                        let is_object_root = match is_object_root(&path) {
+                            Ok(is_object_root) => is_object_root,
+                            Err(e) => return Some(Err(e)),
+                        };
+
+                        if is_object_root {
+                            match self.create_if_matches(&path) {
+                                Some(Ok(inventory)) => return Some(Ok(inventory)),
+                                Some(Err(e)) => return Some(Err(e)),
+                                _ => (),
                             }
-                            Ok(is_root) if !is_root => {
-                                self.dir_iters.push(self.current.replace(None).unwrap());
-                                match fs::read_dir(&path) {
-                                    Ok(next) => {
-                                        self.current.replace(Some(next));
-                                    }
-                                    Err(e) => error!("{:#}", e),
+                        } else {
+                            self.dir_iters.push(self.current.replace(None).unwrap());
+                            match fs::read_dir(&path) {
+                                Ok(next) => {
+                                    self.current.replace(Some(next));
                                 }
+                                Err(e) => return Some(Err(e.into())),
                             }
-                            Err(e) => error!("{:#}", e),
-                            _ => unreachable!(),
                         }
                     }
-                    _ => (),
-                },
+                }
             }
         }
     }
 }
 
+impl FsStorage {
+    pub fn new(storage_root: impl AsRef<Path>) -> Self {
+        Self {
+            storage_root: storage_root.as_ref().to_path_buf(),
+        }
+    }
+}
+
+impl Storage for FsStorage {
+    /// Reads the file at the specified path and writes its contents to the provided sink.
+    fn read<W: Write>(&self, path: &str, sink: &mut W) -> Result<()> {
+        io::copy(&mut File::open(self.storage_root.join(path))?, sink)?;
+        Ok(())
+    }
+
+    /// Lists the contents of the specified directory. If `recursive` is `true`, then all leaf-nodes
+    /// are returned. If the directory does not exist, or is empty, then an empty vector is returned.
+    /// The returned paths are all relative the directory that was listed.
+    fn list(&self, path: &str, recursive: bool) -> Result<Vec<Listing>> {
+        let mut listings = Vec::new();
+        let root = self.storage_root.join(path);
+
+        if fs::metadata(&root).is_err() {
+            return Ok(listings);
+        }
+
+        let mut walker = WalkDir::new(&root);
+
+        if !recursive {
+            walker = walker.max_depth(1);
+        }
+
+        for path in walker {
+            let path = path?;
+
+            let relative_path = util::convert_backslash_to_forward(
+                pathdiff::diff_paths(path.path(), &root)
+                    .unwrap()
+                    .to_string_lossy()
+                    .as_ref(),
+            )
+            .to_string();
+
+            if path.file_type().is_file() {
+                listings.push(Listing::File(Cow::Owned(relative_path)));
+            } else if path.file_type().is_dir() {
+                if path.path() != root.as_path() && (!recursive || util::dir_is_empty(path.path())?)
+                {
+                    listings.push(Listing::Directory(Cow::Owned(relative_path)));
+                }
+            } else {
+                listings.push(Listing::Other(Cow::Owned(relative_path)))
+            }
+        }
+
+        Ok(listings)
+    }
+}
+
 /// Returns true if the path contains an OCFL object root marker file
 fn is_object_root<P: AsRef<Path>>(path: P) -> Result<bool> {
     for entry in fs::read_dir(path)? {
@@ -812,20 +930,6 @@ fn is_object_root<P: AsRef<Path>>(path: P) -> Result<bool> {
     Ok(false)
 }
 
-fn parse_inventory_optional<A, B>(object_root: A, storage_root: B) -> Option<Inventory>
-where
-    A: AsRef<Path>,
-    B: AsRef<Path>,
-{
-    match parse_inventory(object_root, storage_root) {
-        Ok(inventory) => Some(inventory),
-        Err(e) => {
-            error!("{:#}", e);
-            None
-        }
-    }
-}
-
 /// Parses the HEAD inventory of the OCFL object that's rooted in the specified directory.
 /// This is normally the `inventory.json` file in the object's root, but it could also be
 /// the inventory file in an extension directory, such as the mutable HEAD extension.
@@ -862,7 +966,6 @@ where
 fn parse_inventory_file<P: AsRef<Path>>(inventory_file: P) -> Result<Inventory> {
     let bytes = file_to_bytes(inventory_file)?;
     let inventory: Inventory = serde_json::from_slice(&bytes)?;
-    inventory.validate()?;
     Ok(inventory)
 }
 
