@@ -1,23 +1,28 @@
 use core::fmt;
 use std::borrow::Cow;
 use std::cmp::Ordering;
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::convert::{TryFrom, TryInto};
-use std::fmt::{Display, Formatter};
+use std::fmt::{Display, Formatter, Write};
 use std::hash::{Hash, Hasher};
 use std::path;
+use std::path::Path;
 use std::rc::Rc;
 use std::str::{FromStr, Split};
 
 use chrono::{DateTime, Local};
 use once_cell::sync::Lazy;
 use regex::Regex;
-use serde::{Deserialize, Serialize};
+use serde::de::Visitor;
+use serde::{Deserialize, Deserializer, Serialize, Serializer};
+use VersionRef::Head;
 
 use crate::ocfl::bimap::PathBiMap;
+use crate::ocfl::consts::MUTABLE_HEAD_EXT_DIR;
 use crate::ocfl::digest::HexDigest;
 use crate::ocfl::error::{Result, RocflError};
 use crate::ocfl::inventory::{Inventory, Version};
+use crate::ocfl::VersionRef::Number;
 use crate::ocfl::{util, DigestAlgorithm};
 
 static VERSION_REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r#"^v\d+$"#).unwrap());
@@ -31,10 +36,67 @@ pub struct VersionNum {
     pub width: u32,
 }
 
-// TODO model this as LogicalPath and ContentPath -- biggest impact will be on the PathBiMap
-/// Represents a logical or content path.
+/// Represents either a specific version number or whatever the current head version is
+pub enum VersionRef {
+    Number(VersionNum),
+    Head,
+}
+
+pub trait InventoryPath {
+    /// Returns an iterable containing each segment of the path split on the `/` separator
+    fn parts(&self) -> Split<char>;
+
+    /// Returns the parent path of this path.
+    fn parent(&self) -> Self;
+
+    /// Returns the part of the logical path that's after the final `/` or the entire path if
+    /// there is no `/`
+    fn filename(&self) -> &str;
+
+    /// Creates a new path by joining this path with another
+    fn resolve(&self, other: &Self) -> Self;
+
+    /// Returns true if the path ends with the given suffix
+    fn ends_with(&self, suffix: &str) -> bool;
+
+    /// Returns true if the path starts with the given prefix
+    fn starts_with(&self, prefix: &str) -> bool;
+
+    /// Returns a reference to the path represented as a `Path`
+    fn as_path(&self) -> &Path;
+
+    /// Returns a reference to the path represented as a `str`
+    fn as_str(&self) -> &str;
+
+    /// Returns true if the path is empty
+    fn is_empty(&self) -> bool;
+}
+
 #[derive(Deserialize, Serialize, Debug, Eq, Ord, PartialOrd, PartialEq, Hash, Clone)]
-pub struct InventoryPath(String);
+struct InventoryPathInner(String);
+
+/// Represents the logical path to a file in an object.
+#[derive(Serialize, Debug, Eq, Ord, PartialOrd, PartialEq, Hash, Clone)]
+#[serde(transparent)]
+pub struct LogicalPath {
+    inner: InventoryPathInner,
+}
+
+/// Represents a path within a version's content directory. This path must be relative the object
+/// root.
+#[derive(Debug, Eq, Ord, PartialOrd, PartialEq, Hash, Clone)]
+pub struct ContentPath {
+    inner: InventoryPathInner,
+    /// The version the content path exists in. This will be a version number, except in the case
+    /// when the path is in the mutable head extension
+    pub version: ContentPathVersion,
+}
+
+#[derive(Debug, Eq, Ord, PartialOrd, PartialEq, Hash, Copy, Clone)]
+pub enum ContentPathVersion {
+    VersionNum(VersionNum),
+    MutableHead,
+}
 
 /// Represents a version of an OCFL object
 #[derive(Debug, Eq, PartialEq, Clone)]
@@ -48,7 +110,7 @@ pub struct ObjectVersion {
     /// Metadata about the version
     pub version_details: VersionDetails,
     /// A map of files (logical paths) in the version to details about the files.
-    pub state: HashMap<Rc<InventoryPath>, FileDetails>,
+    pub state: HashMap<Rc<LogicalPath>, FileDetails>,
 }
 
 /// Details about a file in an OCFL object
@@ -59,7 +121,7 @@ pub struct FileDetails {
     /// The digest algorithm
     pub digest_algorithm: DigestAlgorithm,
     /// The path to the file relative the object root
-    pub content_path: Rc<InventoryPath>,
+    pub content_path: Rc<ContentPath>,
     /// The path to the file relative the storage root
     pub storage_path: String,
     /// The version metadata for when the file was last updated
@@ -110,24 +172,29 @@ pub struct CommitMeta {
 /// Represents a change to a file
 #[derive(Debug, Eq, PartialEq, Clone)]
 pub enum Diff {
-    Added(Rc<InventoryPath>),
-    Modified(Rc<InventoryPath>),
-    Deleted(Rc<InventoryPath>),
+    Added(Rc<LogicalPath>),
+    Modified(Rc<LogicalPath>),
+    Deleted(Rc<LogicalPath>),
     Renamed {
-        original: Vec<Rc<InventoryPath>>,
-        renamed: Vec<Rc<InventoryPath>>,
+        original: Vec<Rc<LogicalPath>>,
+        renamed: Vec<Rc<LogicalPath>>,
     },
 }
 
+pub(crate) struct PrettyPrintSet<'a, T: Display>(pub(crate) &'a HashSet<T>);
+
 impl VersionNum {
-    /// Creates a new VersionNum with width 0
-    pub fn new(number: u32) -> Self {
-        Self { number, width: 0 }
+    /// Creates a new VersionNum at version 1 with no zero-padding
+    pub fn v1() -> Self {
+        Self {
+            number: 1,
+            width: 0,
+        }
     }
 
-    /// Creates a new VersionNum
-    pub fn with_width(number: u32, width: u32) -> Self {
-        Self { number, width }
+    /// Creates a new VersionNum at version 1 with the specified zero-padding width
+    pub fn v1_with_width(width: u32) -> Self {
+        Self { number: 1, width }
     }
 
     /// Returns the previous version, or an Error if the previous version is invalid (less than 1).
@@ -173,7 +240,7 @@ impl TryFrom<&str> for VersionNum {
     /// the version string is invalid.
     fn try_from(version: &str) -> Result<Self, Self::Error> {
         if !VERSION_REGEX.is_match(version) {
-            return Err(RocflError::IllegalArgument(format!(
+            return Err(RocflError::InvalidValue(format!(
                 "Invalid version {}",
                 version
             )));
@@ -182,7 +249,7 @@ impl TryFrom<&str> for VersionNum {
         match version[1..].parse::<u32>() {
             Ok(num) => {
                 if num < 1 {
-                    return Err(RocflError::IllegalArgument(format!(
+                    return Err(RocflError::InvalidValue(format!(
                         "Invalid version {}",
                         version
                     )));
@@ -198,7 +265,7 @@ impl TryFrom<&str> for VersionNum {
                     width: width as u32,
                 })
             }
-            Err(_) => Err(RocflError::IllegalArgument(format!(
+            Err(_) => Err(RocflError::InvalidValue(format!(
                 "Invalid version {}",
                 version
             ))),
@@ -212,7 +279,7 @@ impl TryFrom<u32> for VersionNum {
     /// Parses a positive integer into a `VersionNum`. An error is returned if it is invalid.
     fn try_from(version: u32) -> Result<Self, Self::Error> {
         if version < 1 {
-            return Err(RocflError::IllegalArgument(format!(
+            return Err(RocflError::InvalidValue(format!(
                 "Invalid version number {}",
                 version
             )));
@@ -235,7 +302,7 @@ impl FromStr for VersionNum {
             Ok(v) => Ok(v),
             Err(_) => match u32::from_str(s) {
                 Ok(parsed) => Ok(VersionNum::try_from(parsed)?),
-                Err(_) => Err(RocflError::IllegalArgument(format!(
+                Err(_) => Err(RocflError::InvalidValue(format!(
                     "Invalid version number {}",
                     s
                 ))),
@@ -282,23 +349,60 @@ impl Ord for VersionNum {
     }
 }
 
-impl InventoryPath {
+impl VersionRef {
+    pub fn resolve(&self, head_num: VersionNum) -> VersionNum {
+        match self {
+            Number(num) => *num,
+            Head => head_num,
+        }
+    }
+}
+
+impl From<VersionNum> for VersionRef {
+    fn from(num: VersionNum) -> Self {
+        Self::Number(num)
+    }
+}
+
+impl From<Option<VersionNum>> for VersionRef {
+    fn from(num: Option<VersionNum>) -> Self {
+        num.map_or(Head, Number)
+    }
+}
+
+impl TryFrom<u32> for VersionRef {
+    type Error = RocflError;
+
+    fn try_from(value: u32) -> Result<Self, Self::Error> {
+        Ok(VersionNum::try_from(value)?.into())
+    }
+}
+
+impl TryFrom<&str> for VersionRef {
+    type Error = RocflError;
+
+    fn try_from(value: &str) -> Result<Self, Self::Error> {
+        Ok(VersionNum::try_from(value)?.into())
+    }
+}
+
+impl InventoryPath for InventoryPathInner {
     /// Returns an iterable containing each segment of the path split on the `/` separator
-    pub fn parts(&self) -> Split<char> {
+    fn parts(&self) -> Split<char> {
         self.0.split('/')
     }
 
     /// Returns the parent path of this path.
-    pub fn parent(&self) -> InventoryPath {
+    fn parent(&self) -> Self {
         match self.0.rfind('/') {
-            Some(last_slash) => InventoryPath(self.0.as_str()[0..last_slash].into()),
-            None => InventoryPath("".to_string()),
+            Some(last_slash) => Self(self.0.as_str()[0..last_slash].into()),
+            None => Self("".to_string()),
         }
     }
 
     /// Returns the part of the logical path that's after the final `/` or the entire path if
     /// there is no `/`
-    pub fn filename(&self) -> &str {
+    fn filename(&self) -> &str {
         match self.0.rfind('/') {
             Some(last_slash) => &self.0.as_str()[last_slash + 1..],
             None => self.0.as_str(),
@@ -306,11 +410,163 @@ impl InventoryPath {
     }
 
     /// Creates a new path by joining this path with another
-    pub fn resolve(&self, other: &InventoryPath) -> InventoryPath {
+    fn resolve(&self, other: &Self) -> Self {
         if self.0.is_empty() {
             other.clone()
         } else {
-            InventoryPath(format!("{}/{}", self.0, other.0))
+            Self(format!("{}/{}", self.0, other.0))
+        }
+    }
+
+    /// Returns true if the path ends with the given suffix
+    fn ends_with(&self, suffix: &str) -> bool {
+        self.0.ends_with(suffix)
+    }
+
+    /// Returns true if the path starts with the given prefix
+    fn starts_with(&self, prefix: &str) -> bool {
+        self.0.starts_with(prefix)
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn as_path(&self) -> &Path {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `str`
+    fn as_str(&self) -> &str {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn is_empty(&self) -> bool {
+        self.0.is_empty()
+    }
+}
+
+impl InventoryPath for LogicalPath {
+    /// Returns an iterable containing each segment of the path split on the `/` separator
+    fn parts(&self) -> Split<char> {
+        self.inner.parts()
+    }
+
+    /// Returns the parent path of this path.
+    fn parent(&self) -> Self {
+        Self {
+            inner: self.inner.parent(),
+        }
+    }
+
+    /// Returns the part of the logical path that's after the final `/` or the entire path if
+    /// there is no `/`
+    fn filename(&self) -> &str {
+        self.inner.filename()
+    }
+
+    /// Creates a new path by joining this path with another
+    fn resolve(&self, other: &Self) -> Self {
+        Self {
+            inner: self.inner.resolve(&other.inner),
+        }
+    }
+
+    /// Returns true if the path ends with the given suffix
+    fn ends_with(&self, suffix: &str) -> bool {
+        self.inner.ends_with(suffix)
+    }
+
+    /// Returns true if the path starts with the given prefix
+    fn starts_with(&self, prefix: &str) -> bool {
+        self.inner.starts_with(prefix)
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn as_path(&self) -> &Path {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `str`
+    fn as_str(&self) -> &str {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn is_empty(&self) -> bool {
+        self.inner.is_empty()
+    }
+}
+
+impl InventoryPath for ContentPath {
+    /// Returns an iterable containing each segment of the path split on the `/` separator
+    fn parts(&self) -> Split<char> {
+        self.inner.parts()
+    }
+
+    /// Returns the parent path of this path.
+    fn parent(&self) -> Self {
+        Self {
+            inner: self.inner.parent(),
+            version: self.version,
+        }
+    }
+
+    /// Returns the part of the logical path that's after the final `/` or the entire path if
+    /// there is no `/`
+    fn filename(&self) -> &str {
+        self.inner.filename()
+    }
+
+    /// Creates a new path by joining this path with another
+    fn resolve(&self, other: &Self) -> Self {
+        Self {
+            inner: self.inner.resolve(&other.inner),
+            version: self.version,
+        }
+    }
+
+    /// Returns true if the path ends with the given suffix
+    fn ends_with(&self, suffix: &str) -> bool {
+        self.inner.ends_with(suffix)
+    }
+
+    /// Returns true if the path starts with the given prefix
+    fn starts_with(&self, prefix: &str) -> bool {
+        self.inner.starts_with(prefix)
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn as_path(&self) -> &Path {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `str`
+    fn as_str(&self) -> &str {
+        self.as_ref()
+    }
+
+    /// Returns a reference to the path represented as a `Path`
+    fn is_empty(&self) -> bool {
+        self.inner.is_empty()
+    }
+}
+
+impl LogicalPath {
+    /// Creates a new content path by directly mapping the logical path to a content path
+    pub fn to_content_path(&self, version_num: VersionNum, content_dir: &str) -> ContentPath {
+        ContentPath::for_logical_path(version_num, content_dir, self)
+    }
+}
+
+impl ContentPath {
+    /// Creates a new content path by directly mapping the logical path to a content path
+    pub fn for_logical_path(
+        version_num: VersionNum,
+        content_dir: &str,
+        logical_path: &LogicalPath,
+    ) -> Self {
+        Self {
+            inner: InventoryPathInner(format!("{}/{}/{}", version_num, content_dir, logical_path)),
+            version: ContentPathVersion::VersionNum(version_num),
         }
     }
 }
@@ -318,10 +574,12 @@ impl InventoryPath {
 // It looks like its not possible to implement `impl<T: AsRef<str> TryFrom<t>`
 // https://github.com/rust-lang/rust/issues/50133
 
-impl TryFrom<&str> for InventoryPath {
+impl TryFrom<&str> for InventoryPathInner {
     type Error = RocflError;
 
     fn try_from(value: &str) -> Result<Self, Self::Error> {
+        // Paths with leading or trailing slashes are technically illegal, but stripping here makes
+        // the code easier to deal with.
         let trimmed = value.trim_start_matches('/').trim_end_matches('/');
 
         if !trimmed.is_empty() {
@@ -330,7 +588,7 @@ impl TryFrom<&str> for InventoryPath {
                 .any(|part| part == "." || part == ".." || part.is_empty());
 
             if has_illegal_part {
-                return Err(RocflError::IllegalArgument(format!(
+                return Err(RocflError::InvalidValue(format!(
                     "Paths may not contain '.', '..', or '' parts. Found: {} ",
                     value
                 )));
@@ -341,7 +599,42 @@ impl TryFrom<&str> for InventoryPath {
     }
 }
 
-impl TryFrom<String> for InventoryPath {
+impl TryFrom<&str> for LogicalPath {
+    type Error = RocflError;
+
+    fn try_from(value: &str) -> Result<Self, Self::Error> {
+        Ok(Self {
+            inner: InventoryPathInner::try_from(value)?,
+        })
+    }
+}
+
+impl TryFrom<&str> for ContentPath {
+    type Error = RocflError;
+
+    fn try_from(value: &str) -> Result<Self, Self::Error> {
+        let inner = InventoryPathInner::try_from(value)?;
+
+        // Mutable head paths do not have a version
+        let version = if value.starts_with(MUTABLE_HEAD_EXT_DIR) {
+            ContentPathVersion::MutableHead
+        } else {
+            match value.find('/') {
+                Some(index) => ContentPathVersion::VersionNum(value[0..index].try_into()?),
+                None => {
+                    return Err(RocflError::InvalidValue(format!(
+                        "Content paths must begin with a valid version number. Found: {} ",
+                        value
+                    )));
+                }
+            }
+        };
+
+        Ok(Self { inner, version })
+    }
+}
+
+impl TryFrom<String> for InventoryPathInner {
     type Error = RocflError;
 
     fn try_from(value: String) -> Result<Self, Self::Error> {
@@ -349,7 +642,25 @@ impl TryFrom<String> for InventoryPath {
     }
 }
 
-impl TryFrom<&String> for InventoryPath {
+impl TryFrom<String> for LogicalPath {
+    type Error = RocflError;
+
+    fn try_from(value: String) -> Result<Self, Self::Error> {
+        Ok(Self {
+            inner: InventoryPathInner::try_from(value)?,
+        })
+    }
+}
+
+impl TryFrom<String> for ContentPath {
+    type Error = RocflError;
+
+    fn try_from(value: String) -> Result<Self, Self::Error> {
+        Self::try_from(value.as_str())
+    }
+}
+
+impl TryFrom<&String> for InventoryPathInner {
     type Error = RocflError;
 
     fn try_from(value: &String) -> Result<Self, Self::Error> {
@@ -357,7 +668,25 @@ impl TryFrom<&String> for InventoryPath {
     }
 }
 
-impl TryFrom<Cow<'_, str>> for InventoryPath {
+impl TryFrom<&String> for LogicalPath {
+    type Error = RocflError;
+
+    fn try_from(value: &String) -> Result<Self, Self::Error> {
+        Ok(Self {
+            inner: InventoryPathInner::try_from(value)?,
+        })
+    }
+}
+
+impl TryFrom<&String> for ContentPath {
+    type Error = RocflError;
+
+    fn try_from(value: &String) -> Result<Self, Self::Error> {
+        Self::try_from(value.as_str())
+    }
+}
+
+impl TryFrom<Cow<'_, str>> for InventoryPathInner {
     type Error = RocflError;
 
     fn try_from(value: Cow<'_, str>) -> Result<Self, Self::Error> {
@@ -365,37 +694,183 @@ impl TryFrom<Cow<'_, str>> for InventoryPath {
     }
 }
 
-impl From<InventoryPath> for String {
-    fn from(path: InventoryPath) -> Self {
+impl TryFrom<Cow<'_, str>> for LogicalPath {
+    type Error = RocflError;
+
+    fn try_from(value: Cow<'_, str>) -> Result<Self, Self::Error> {
+        Ok(Self {
+            inner: InventoryPathInner::try_from(value)?,
+        })
+    }
+}
+
+impl TryFrom<Cow<'_, str>> for ContentPath {
+    type Error = RocflError;
+
+    fn try_from(value: Cow<'_, str>) -> Result<Self, Self::Error> {
+        Self::try_from(value.as_ref())
+    }
+}
+
+impl From<InventoryPathInner> for String {
+    fn from(path: InventoryPathInner) -> Self {
         path.0
     }
 }
 
-impl AsRef<String> for InventoryPath {
-    fn as_ref(&self) -> &String {
+impl From<LogicalPath> for String {
+    fn from(path: LogicalPath) -> Self {
+        path.inner.0
+    }
+}
+
+impl From<ContentPath> for String {
+    fn from(path: ContentPath) -> Self {
+        path.inner.0
+    }
+}
+
+impl AsRef<str> for InventoryPathInner {
+    fn as_ref(&self) -> &str {
         &self.0
     }
 }
 
-impl Display for InventoryPath {
+impl AsRef<str> for LogicalPath {
+    fn as_ref(&self) -> &str {
+        self.inner.as_ref()
+    }
+}
+
+impl AsRef<str> for ContentPath {
+    fn as_ref(&self) -> &str {
+        self.inner.as_ref()
+    }
+}
+
+impl AsRef<Path> for InventoryPathInner {
+    fn as_ref(&self) -> &Path {
+        self.0.as_ref()
+    }
+}
+
+impl AsRef<Path> for LogicalPath {
+    fn as_ref(&self) -> &Path {
+        self.inner.as_ref()
+    }
+}
+
+impl AsRef<Path> for ContentPath {
+    fn as_ref(&self) -> &Path {
+        self.inner.as_ref()
+    }
+}
+
+impl Display for InventoryPathInner {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         write!(f, "{}", self.0)
     }
 }
 
+impl Display for LogicalPath {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        self.inner.fmt(f)
+    }
+}
+
+impl Display for ContentPath {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        self.inner.fmt(f)
+    }
+}
+
+impl Serialize for ContentPath {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        serializer.serialize_str(self.as_str())
+    }
+}
+
+impl<'de> Deserialize<'de> for ContentPath {
+    fn deserialize<D>(deserializer: D) -> Result<ContentPath, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        struct ContentPathVisitor;
+
+        impl<'de> Visitor<'de> for ContentPathVisitor {
+            type Value = ContentPath;
+
+            fn expecting(&self, formatter: &mut Formatter) -> std::fmt::Result {
+                formatter.write_str("a path string that is a valid OCFL content path")
+            }
+
+            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                if v.starts_with('/') || v.ends_with('/') {
+                    return Err(E::custom(format!(
+                        "Invalid value: content paths may not begin or end with a '/'. Found: {}",
+                        v
+                    )));
+                }
+
+                v.try_into()
+                    .map_err(|e: RocflError| E::custom(e.to_string()))
+            }
+        }
+
+        deserializer.deserialize_str(ContentPathVisitor)
+    }
+}
+
+impl<'de> Deserialize<'de> for LogicalPath {
+    fn deserialize<D>(deserializer: D) -> Result<LogicalPath, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        struct LogicalPathVisitor;
+
+        impl<'de> Visitor<'de> for LogicalPathVisitor {
+            type Value = LogicalPath;
+
+            fn expecting(&self, formatter: &mut Formatter) -> std::fmt::Result {
+                formatter.write_str("a path string that is a valid OCFL logical path")
+            }
+
+            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
+            where
+                E: serde::de::Error,
+            {
+                if v.starts_with('/') || v.ends_with('/') {
+                    return Err(E::custom(format!(
+                        "Invalid value: logical paths may not begin or end with a '/'. Found: {}",
+                        v
+                    )));
+                }
+
+                v.try_into()
+                    .map_err(|e: RocflError| E::custom(e.to_string()))
+            }
+        }
+
+        deserializer.deserialize_str(LogicalPathVisitor)
+    }
+}
+
 impl ObjectVersion {
     /// Creates an `ObjectVersion` by consuming the supplied `Inventory`.
     pub fn from_inventory<S: AsRef<str> + Copy>(
         mut inventory: Inventory,
-        version_num: Option<VersionNum>,
+        version_num: VersionRef,
         object_storage_path: S,
         object_staging_path: Option<S>,
         use_backslashes: bool,
     ) -> Result<Self> {
-        let version_num = match version_num {
-            Some(version) => version,
-            None => inventory.head,
-        };
+        let version_num = version_num.resolve(inventory.head);
 
         let version = inventory.get_version(version_num)?;
         let version_details = VersionDetails::new(version_num, version);
@@ -423,7 +898,7 @@ impl ObjectVersion {
         object_storage_path: S,
         object_staging_path: Option<S>,
         use_backslashes: bool,
-    ) -> Result<HashMap<Rc<InventoryPath>, FileDetails>> {
+    ) -> Result<HashMap<Rc<LogicalPath>, FileDetails>> {
         let mut state = HashMap::new();
 
         let mut current_version_num = target;
@@ -449,12 +924,12 @@ impl ObjectVersion {
                 for (target_path, target_digest) in target_path_map {
                     let content_path = inventory.content_path_for_digest(
                         &target_digest,
-                        Some(current_version_num),
+                        current_version_num.into(),
                         Some(&target_path),
                     )?;
 
                     let storage_path = ObjectVersion::storage_path(
-                        content_path.as_ref().as_ref(),
+                        content_path.as_str(),
                         object_storage_path,
                         use_backslashes,
                         &staging_version_prefix,
@@ -486,12 +961,12 @@ impl ObjectVersion {
                 if entry.is_none() || entry.unwrap().1 != target_digest {
                     let content_path = inventory.content_path_for_digest(
                         &target_digest,
-                        Some(current_version_num),
+                        current_version_num.into(),
                         Some(&target_path),
                     )?;
 
                     let storage_path = ObjectVersion::storage_path(
-                        content_path.as_ref().as_ref(),
+                        content_path.as_str(),
                         object_storage_path,
                         use_backslashes,
                         &staging_version_prefix,
@@ -553,7 +1028,7 @@ impl ObjectVersion {
 
 impl FileDetails {
     pub fn new(
-        content_path: Rc<InventoryPath>,
+        content_path: Rc<ContentPath>,
         storage_path: String,
         digest: Rc<HexDigest>,
         digest_algorithm: DigestAlgorithm,
@@ -605,14 +1080,8 @@ impl VersionDetails {
 
 impl ObjectVersionDetails {
     /// Creates `ObjectVersionDetails` by consuming the `Inventory`.
-    pub fn from_inventory(
-        mut inventory: Inventory,
-        version_num: Option<VersionNum>,
-    ) -> Result<Self> {
-        let version_num = match version_num {
-            Some(version) => version,
-            None => inventory.head,
-        };
+    pub fn from_inventory(mut inventory: Inventory, version_num: VersionRef) -> Result<Self> {
+        let version_num = version_num.resolve(inventory.head);
 
         let version = inventory.remove_version(version_num)?;
         let version_details = VersionDetails::from_version(version_num, version);
@@ -646,7 +1115,7 @@ impl CommitMeta {
     /// Sets the commit user. `name` must be provided if `address` is provided.
     pub fn with_user(mut self, name: Option<String>, address: Option<String>) -> Result<Self> {
         if address.is_some() && name.is_none() {
-            return Err(RocflError::IllegalArgument(
+            return Err(RocflError::InvalidValue(
                 "User name must be set when user address is set.".to_string(),
             ));
         }
@@ -671,7 +1140,7 @@ impl CommitMeta {
 impl Diff {
     /// This method returns the path associated with the diff. If there are multiple paths,
     /// it is the first path on the left hand side.
-    pub fn path(&self) -> &Rc<InventoryPath> {
+    pub fn path(&self) -> &Rc<LogicalPath> {
         match self {
             Diff::Added(path) => path,
             Diff::Modified(path) => path,
@@ -683,6 +1152,20 @@ impl Diff {
     }
 }
 
+impl<'a, T: Display> Display for PrettyPrintSet<'a, T> {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        f.write_char('[')?;
+        let max = self.0.len() - 1;
+        for (i, entry) in self.0.iter().enumerate() {
+            write!(f, "{}", entry)?;
+            if i < max {
+                write!(f, ", ")?;
+            }
+        }
+        f.write_char(']')
+    }
+}
+
 /// Joins to strings using the file system separator
 fn join(use_backslashes: bool, parent: &str, child: &str) -> String {
     if use_backslashes {
@@ -704,48 +1187,48 @@ fn convert_path_separator(use_backslashes: bool, path: String) -> String {
 mod tests {
     use std::convert::{TryFrom, TryInto};
 
-    use crate::ocfl::InventoryPath;
+    use crate::ocfl::LogicalPath;
 
     #[test]
     fn create_logical_path_when_valid() {
         let value = "foo/.bar/baz.txt";
-        let path: InventoryPath = value.try_into().unwrap();
-        assert_eq!(value, path.0);
+        let path: LogicalPath = value.try_into().unwrap();
+        assert_eq!(value, path.inner.0);
     }
 
     #[test]
     fn create_logical_path_when_root() {
-        let path: InventoryPath = "/".try_into().unwrap();
-        assert_eq!("", path.0);
+        let path: LogicalPath = "/".try_into().unwrap();
+        assert_eq!("", path.inner.0);
     }
 
     #[test]
     fn remove_leading_and_trailing_slashes_from_logical_paths() {
-        let path: InventoryPath = "//foo/bar/baz//".try_into().unwrap();
-        assert_eq!("foo/bar/baz", path.0);
+        let path: LogicalPath = "//foo/bar/baz//".try_into().unwrap();
+        assert_eq!("foo/bar/baz", path.inner.0);
     }
 
     #[test]
     #[should_panic(expected = "Paths may not contain")]
     fn reject_logical_paths_with_empty_parts() {
-        InventoryPath::try_from("foo//bar/baz").unwrap();
+        LogicalPath::try_from("foo//bar/baz").unwrap();
     }
 
     #[test]
     #[should_panic(expected = "Paths may not contain")]
     fn reject_logical_paths_with_single_dot() {
-        InventoryPath::try_from("foo/bar/./baz").unwrap();
+        LogicalPath::try_from("foo/bar/./baz").unwrap();
     }
 
     #[test]
     #[should_panic(expected = "Paths may not contain")]
     fn reject_logical_paths_with_double_dot() {
-        InventoryPath::try_from("foo/bar/../baz").unwrap();
+        LogicalPath::try_from("foo/bar/../baz").unwrap();
     }
 
     #[test]
     #[should_panic(expected = "Paths may not contain")]
     fn reject_logical_paths_with_double_dot_leading() {
-        InventoryPath::try_from("../foo/bar/baz").unwrap();
+        LogicalPath::try_from("../foo/bar/baz").unwrap();
     }
 }
